import cloud_info_provider.providers.base
import mock
from cloud_info_provider import glue
from cloud_info_provider.tests import base
from cloud_info_provider.tests import utils as utils
from cloud_info_provider.tests.data import DATA


class FakeBaseProvider(cloud_info_provider.providers.base.BaseProvider):
    def _load_site_config(self, site_config):
        self.site_config = DATA.site_config


class BaseProviderTest(base.TestCase):
    def setUp(self):
        class Opts(object):
            debug = None
            timeout = 1234
            site_config = None
            insecure = None

        super().setUp()
        self.provider = FakeBaseProvider(Opts())

    def test_provider_build_service(self):
        with mock.patch(
            "cloud_info_provider.providers.utils.find_in_gocdb"
        ) as m_goc_find:
            m_goc_find.return_value = {"foo": "bar"}
            service = {
                "site_name": DATA.site_name,
                "name": f"Cloud Compute service at {DATA.site_name}",
                "other_info": {"foo": "bar"},
                "associations": {"AdminDomain": [DATA.site_name]},
                "status_info": f"https://argo.egi.eu/egi/report-status/Critical/SITES/{DATA.site_name}",
                "capability": [
                    "executionmanagement.dynamicvmdeploy",
                    "security.accounting",
                ],
                "type": "org.cloud.iaas",
                "id": "service",
                "quality_level": "production",
                "service_aup": "http://go.egi.eu/aup",
            }
            assert utils.compare_glue(service, self.provider.build_service())

    def test_provider_build_manager(self):
        self.provider.add_glue(glue.CloudComputingService(id="foo", status_info="bar"))
        manager = {"id": "manager", "associations": {"CloudComputingService": ["foo"]}}
        assert utils.compare_glue(manager, self.provider.build_manager())

    def test_provider_build_endpoint(self):
        with mock.patch(
            "cloud_info_provider.providers.utils.get_endpoint_ca_information"
        ) as m_ca_get:
            m_ca_get.return_value = {"issuer": "foo_ca", "trusted_cas": ["ca1", "ca2"]}
            self.provider.add_glue(
                glue.CloudComputingService(id="foo", status_info="bar")
            )
            endpoint = {
                "id": "endpoint",
                "name": "Cloud computing endpoint for endpoint",
                "url": "https://foo.example.org:5000/v3",
                "associations": {"CloudComputingService": ["foo"]},
                "capability": [],
                "quality_level": "production",
                "serving_state": "production",
                "interface_name": "",
                "health_state": "ok",
                "health_state_info": "Endpoint funtioning properly",
                "technology": "webservice",
                "downtime_info": "https://goc.egi.eu/portal/index.php?Page_Type=Downtimes_Calendar&site=SITE_NAME",
                "issuer_ca": "foo_ca",
                "trusted_cas": ["ca1", "ca2"],
            }
            assert utils.compare_glue(endpoint, self.provider.build_endpoint())

    def test_provider_build_shares(self):
        self.provider.build_shares()
        self.provider.get_objs("Share") == []

    def test_get_goc_info_no_svc(self):
        assert self.provider._get_goc_info("baz") == {}

    def test_get_goc_info(self):
        self.provider.goc_service_type = "svc"
        with mock.patch(
            "cloud_info_provider.providers.utils.find_in_gocdb"
        ) as m_goc_find:
            m_goc_find.return_value = {"foo": "bar"}
            info = self.provider._get_goc_info("baz")
            assert self.provider._goc_info == {"baz": {"foo": "bar"}}
            assert info == {"foo": "bar"}

    def test_fetch(self):
        with utils.nested(
            mock.patch("cloud_info_provider.providers.utils.find_in_gocdb"),
            mock.patch(
                "cloud_info_provider.providers.utils.get_endpoint_ca_information"
            ),
        ) as (m_goc_find, m_ca_get):
            m_goc_find.return_value = {"foo": "bar"}
            m_ca_get.return_value = {"issuer": "foo_ca", "trusted_cas": ["ca1", "ca2"]}
            self.provider.fetch()
            # just check the complexity here
            assert self.provider.service.complexity == "endpointType=1,share=0"
            assert self.provider.manager
            assert self.provider.endpoint



class BasePublisher(object):
    def __init__(self, opts):
        self.opts = opts

    @staticmethod
    def populate_parser(parser):
        """Populate the argparser 'parser' with the needed options."""
        pass

    def publish(self, output):
        raise NotImplementedError



"""
GlueSchema 2.1 Objects
"""

import datetime
from enum import Enum
from typing import Optional

from pydantic import BaseModel


class BoolEnum(Enum):
    """A class to deal with true/false or not knwon"""

    TRUE = True
    FALSE = False
    UNKNOWN = "UNKNOWN"


class GlueBase(BaseModel):
    id: str
    name: Optional[str] = None
    creation_time: datetime.datetime = datetime.datetime.now(datetime.timezone.utc)
    # 12 hours validity
    validity: int = 3600 * 12
    other_info: dict = {}
    associations: dict = {}

    def add_association(self, name, value):
        assoc = self.associations.get(name, [])
        assoc.append(value)
        self.associations[name] = assoc

    def add_associated_object(self, obj):
        class_name = obj.__class__.__name__
        self.add_association(class_name, obj.id)


class CloudComputingService(GlueBase):
    type: str = "org.cloud.iaas"
    quality_level: str = "production"
    status_info: str
    service_aup: str = "http://go.egi.eu/aup"
    complexity: Optional[str] = None
    capability: list[str] = [
        "executionmanagement.dynamicvmdeploy",
        "security.accounting",
    ]
    total_vm: Optional[int] = None
    running_vm: Optional[int] = None
    suspended_vm: Optional[int] = None
    halted_vm: Optional[int] = None


class CloudComputingManager(GlueBase):
    product_name: Optional[str] = None
    product_version: Optional[str] = None
    hypervisor_name: Optional[str] = None
    hypervisor_version: Optional[str] = None
    total_cpus: Optional[int] = None
    total_ram: Optional[int] = None
    instance_max_cpu: Optional[int] = None
    instance_min_cpu: Optional[int] = None
    instance_max_ram: Optional[int] = None
    instance_min_ram: Optional[int] = None
    network_virtualization_type: Optional[str] = None
    cpu_virtualization_type: Optional[str] = None
    virtual_disk_format: Optional[str] = None
    failover: Optional[BoolEnum] = None
    live_migration: Optional[BoolEnum] = None
    vm_backup_restore: Optional[BoolEnum] = None


class CloudComputingEndpoint(GlueBase):
    url: str
    capability: list[str] = []
    quality_level: str = "production"
    serving_state: str = "production"
    interface_name: str
    interface_version: Optional[str] = None
    # FIXME: This should be actually computed
    health_state: str = "ok"
    health_state_info: Optional[str] = None
    technology: str = "webservice"
    implementor: Optional[str] = None
    implementation_name: Optional[str] = None
    implementation_version: Optional[str] = None
    downtime_info: Optional[str] = None
    semantics: Optional[str] = None
    authentication: Optional[str] = None
    issuer_ca: Optional[str] = None
    trusted_cas: Optional[list[str]] = None


class CloudComputingImage(GlueBase):
    marketplace_url: Optional[str] = None
    osPlatform: str = ""
    osName: str = ""
    osVersion: Optional[str] = None
    description: Optional[str] = None
    access_info: str = "none"


class CloudComputingInstanceType(GlueBase):
    platform: str = "UNKNOWN"
    cpu: int = 0
    ram: int = 0
    disk: int = 0
    network_in: BoolEnum = BoolEnum.UNKNOWN
    network_out: BoolEnum = BoolEnum.TRUE
    network_info: Optional[str] = None


class CloudComputingVirtualAccelerator(GlueBase):
    type: str
    number: int = 0
    vendor: Optional[str] = None
    model: Optional[str] = None
    version: Optional[str] = None
    clock_speed: Optional[int] = None
    memory: Optional[int] = None
    compute_capability: list[str] = []
    virtualization_type: Optional[str] = None


class Policy(GlueBase):
    scheme: str = "org.glite.standard"
    rule: list[str] = []


# These two just have different names, but nothing else
class MappingPolicy(Policy):
    pass


class AccessPolicy(Policy):
    pass


class Share(GlueBase):
    instance_max_cpu: Optional[int] = None
    instance_max_ram: Optional[int] = None
    instance_min_cpu: Optional[int] = None
    instance_min_ram: Optional[int] = None
    sla: Optional[str] = None
    total_vm: Optional[int] = None
    running_vm: Optional[int] = None
    suspended_vm: Optional[int] = None
    halted_vm: Optional[int] = None
    max_vm: Optional[int] = None
    project_id: Optional[str] = None
    network_info: Optional[str] = None
    default_network_type: Optional[str] = None
    public_network_name: Optional[str] = None



"""
Simple utilities for getting information from endpoints
"""

import logging
import socket
from urllib.parse import urlparse
from xml.etree.ElementTree import ParseError  # nosec

import defusedxml.ElementTree
import requests
from OpenSSL import SSL

logger = logging.getLogger(__name__)


def _are_url_similar(u1, u2):
    """Compares 2 urls just considering host and path"""
    if u1[0:2] == u2[0:2]:
        # consider path/ == path which is not strictly the same
        # but good enough for our case
        p1 = u1[2].strip("/")
        p2 = u2[2].strip("/")
        if p1 == p2:
            return True
    return False


def _find_url_in_result(svc_url, result):
    """Finds service matching URL in GOCDB

    Parses XML gotten from GOC and tries to find the service which
    mathces the URL of the service. It returns a dictionary with keys
    `gocdb_id` which has the ID in GOCDB of the service and `site_name`
    with the name of the site. If no match is found, an empty dictionary
    is returned.
    """

    svc_url = urlparse(svc_url)
    for svc in result:
        urls = []
        try:
            urls.append(urlparse(svc.find("URL").text))
        except AttributeError:
            # something is wrong at GOCDB, do not care
            pass
        for ep in svc.find("ENDPOINTS"):
            try:
                urls.append(urlparse(ep.find("URL").text))
            except AttributeError:
                # something is wrong at GOCDB, do not care
                pass
        for url in urls:
            if _are_url_similar(svc_url, url):
                return {
                    "gocdb_id": svc.attrib["PRIMARY_KEY"],
                    "site_name": svc.find("SITENAME").text,
                }
    logger.warning("Unable to find URL %s in GOCDB!", svc_url)
    return {}


def find_in_gocdb(svc_url, svc_type, insecure=False, timeout=None):
    """Find service matching URL and service type in GOCDB"""

    verify = not insecure
    r = requests.get(
        "https://goc.egi.eu/gocdbpi/public/",
        params={"method": "get_service", "service_type": svc_type},
        verify=verify,
        timeout=timeout,
    )
    if r.status_code != 200:
        logger.warning("Something went wrong with GOC %s", r.text)
        return {}
    try:
        goc_xml = defusedxml.ElementTree.fromstring(r.text)
        return _find_url_in_result(svc_url, goc_xml)
    except ParseError:
        logger.warning("Something went wrong with parsing GOC output")
        return {}


def get_dn(x509name):
    """Return the DN of an X509Name object."""
    # str(X509Name) is something like <X509Name object 'DN'>
    obj_name = str(x509name)
    start = obj_name.find("'") + 1
    end = obj_name.rfind("'")
    return obj_name[start:end]


def get_endpoint_ca_information(endpoint_url, insecure=False):
    """Return certificate issuer and trusted CAs list of HTTPS endpoint."""
    ca_info = {"issuer": "UNKNOWN", "trusted_cas": ["UNKNOWN"]}

    verify = SSL.VERIFY_NONE if insecure else SSL.VERIFY_PEER

    scheme = urlparse(endpoint_url).scheme
    host = urlparse(endpoint_url).hostname
    port = urlparse(endpoint_url).port

    if scheme != "https":
        return ca_info

    # use default port for https
    if not port:
        port = 443

    try:
        ctx = SSL.Context(SSL.TLSv1_2_METHOD)
        ctx.set_options(SSL.OP_NO_SSLv2)
        ctx.set_options(SSL.OP_NO_SSLv3)
        ctx.set_verify(verify, lambda conn, cert, errno, depth, ok: ok)
        ctx.set_default_verify_paths()

        client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        client.connect((host, port))

        client_ssl = SSL.Connection(ctx, client)
        client_ssl.set_connect_state()
        client_ssl.set_tlsext_host_name(host.encode("utf-8"))
        client_ssl.do_handshake()

        cert = client_ssl.get_peer_certificate()
        issuer = get_dn(cert.get_issuer())

        client_ca_list = client_ssl.get_client_ca_list()
        trusted_cas = [get_dn(ca) for ca in client_ca_list]

        client_ssl.shutdown()
        client_ssl.close()

        ca_info["issuer"] = issuer
        ca_info["trusted_cas"] = trusted_cas
    except SSL.Error as e:
        logger.warning("Issue when getting CA info from endpoint: %s", e)
    except TimeoutError as e:
        logger.warning("Timeout when getting CA info from endpoint: %s", e)

    return ca_info



import logging

logger = logging.getLogger(__name__)


class CloudInfoException(Exception):
    msg_fmt = "An unknown exception occurred."

    def __init__(self, message=None, **kwargs):
        self.kwargs = kwargs

        if not message:
            try:
                message = self.msg_fmt % kwargs
            except Exception:
                # kwargs doesn't match a variable in the message
                # log the issue and the kwargs
                logger.exception("Exception in string format operation")
                for name, value in kwargs.items():
                    logger.error("%s: %s" % (name, value))
                raise

        super().__init__(message)


class OpenStackProviderException(CloudInfoException):
    pass



import argparse
import logging

import cloud_info_provider
from stevedore import driver, extension


def get_providers():
    def _handle_exception(*args):
        mgr, entry_point, exception = args
        logging.getLogger("stevedore.extension").error(
            (("Cannot load '%s': %s") % (entry_point, exception))
        )

    mgr = extension.ExtensionManager(
        namespace="cip.providers",
        on_load_failure_callback=_handle_exception,
        propagate_map_exceptions=True,
    )
    return dict((x.name, x.plugin) for x in mgr)


def get_formatters():
    mgr = extension.ExtensionManager(
        namespace="cip.formatters",
    )
    return [x.name for x in mgr]


def get_publishers():
    mgr = extension.ExtensionManager(
        namespace="cip.publishers",
    )
    return dict((x.name, x.plugin) for x in mgr)


def get_parser(providers, formatters, publishers):
    parser = argparse.ArgumentParser(
        description="Cloud Information System provider",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        fromfile_prefix_chars="@",
        conflict_handler="resolve",
    )

    parser.add_argument(
        "--version", action="version", version=f"{cloud_info_provider.__version__}"
    )

    parser.add_argument(
        "--middleware",
        metavar="MIDDLEWARE",
        choices=providers,
        default="openstack",
        help=(
            "Middleware used. Only the following middlewares are "
            "supported: {%(choices)s}"
        ),
    )

    parser.add_argument(
        "--format",
        metavar="FORMAT",
        choices=formatters,
        default="glue21json",
        help=("Selects the output format. Allowed values: {%(choices)s}"),
    )

    parser.add_argument(
        "--publisher",
        metavar="PUBLISHER",
        choices=publishers,
        default="stdout",
        help=("Selects where to publish output to. Allowed values: " "%(choices)s}"),
    )

    parser.add_argument(
        "--timeout",
        default=600,
        metavar="<seconds>",
        help="Set request timeout (in seconds).",
    )

    parser.add_argument(
        "--insecure",
        action="store_true",
        help="Explicitly allow to perform 'insecure' "
        "SSL (https) requests. The server's certificate will "
        "not be verified against any certificate authorities. "
        "This option should be used with caution.",
    )

    parser.add_argument(
        "--debug",
        action="store_true",
        default=False,
        help="Provide extra logging information",
    )

    parser.add_argument(
        "--exit-on-share-errors",
        action="store_true",
        default=False,
        help=("Exit if there are errors getting information from a share."),
    )

    for provider_name, provider in providers.items():
        # Do not pass an argument group to plugins that may create their own
        # argument groups (for example keystoneauth loading.session). Passing
        # the root parser avoids nested argument groups which argparse
        # disallows (raises ValueError: argument groups cannot be nested).
        provider.populate_parser(parser)

    for publisher_name, publisher in publishers.items():
        # Likewise, pass the root parser to publishers.
        publisher.populate_parser(parser)

    return parser


def main():
    providers = get_providers()
    formatters = get_formatters()
    publishers = get_publishers()

    opts = get_parser(providers, formatters, publishers).parse_args()

    formatter = driver.DriverManager(
        namespace="cip.formatters",
        name=opts.format,
        invoke_on_load=True,
    )

    publisher = driver.DriverManager(
        namespace="cip.publishers",
        name=opts.publisher,
        invoke_on_load=True,
        invoke_args=(opts,),
    )

    provider_class = providers.get(opts.middleware)
    glue = provider_class(opts).fetch()

    output = formatter.driver.format(opts, glue)
    publisher.driver.publish(output)


if __name__ == "__main__":
    main()



import json
import os


def env(*args, **kwargs):
    """Returns the first environment variable set.

    if none are non-empty, defaults to '' or keyword arg default
    """
    for arg in args:
        value = os.environ.get(arg, None)
        if value:
            return value
    return kwargs.get("default", "")


def pythonize_network_info(network_info):
    """Pythonize Ruby dict-like network_info string. Do nothing otherwise."""
    try:
        return json.loads(network_info.replace(':"', '"').replace("=>", ":"))
    except AttributeError:
        return network_info



import mock
from cloud_info_provider.providers import utils
from cloud_info_provider.tests import base

sample_goc_response = """<?xml version="1.0" encoding="UTF-8"?>
<results>
    <SERVICE_ENDPOINT PRIMARY_KEY="1234G0">
        <PRIMARY_KEY>1234G0</PRIMARY_KEY>
        <HOSTNAME>simple.example.com</HOSTNAME>
        <SERVICE_TYPE>org.openstack.nova</SERVICE_TYPE>
        <CORE></CORE>
        <IN_PRODUCTION>Y</IN_PRODUCTION>
        <NODE_MONITORED>Y</NODE_MONITORED>
        <NOTIFICATIONS>Y</NOTIFICATIONS>
        <SITENAME>FOO-BAR-SITE</SITENAME>
        <COUNTRY_NAME>Spain</COUNTRY_NAME>
        <COUNTRY_CODE>ES</COUNTRY_CODE>
        <ROC_NAME>NGI_IBERGRID</ROC_NAME>
        <URL>https://keystone.example.com:5000/v2.0/?image=18d99a06-c3e5-4157-a0e3-37ec34bdfc24&amp;resource=m1.tiny</URL>
        <ENDPOINTS/>
  </SERVICE_ENDPOINT>
</results>"""

sample_goc_ep_response = """<?xml version="1.0" encoding="UTF-8"?>
<results>
  <SERVICE_ENDPOINT PRIMARY_KEY="00000G0">
    <PRIMARY_KEY>00000G0</PRIMARY_KEY>
    <HOSTNAME>www.baz.example.com</HOSTNAME>
    <GOCDB_PORTAL_URL>https://goc.egi.eu/portal/index.php?Page_Type=Service&amp;id=14541</GOCDB_PORTAL_URL>
    <BETA>N</BETA>
    <SERVICE_TYPE>org.openstack.nova</SERVICE_TYPE>
    <CORE/>
    <IN_PRODUCTION>Y</IN_PRODUCTION>
    <NODE_MONITORED>Y</NODE_MONITORED>
    <NOTIFICATIONS>Y</NOTIFICATIONS>
    <SITENAME>BAR-FOO-SITE</SITENAME>
    <COUNTRY_NAME>Ireland</COUNTRY_NAME>
    <COUNTRY_CODE>IE</COUNTRY_CODE>
    <ROC_NAME>NGI_IE</ROC_NAME>
    <URL/>
    <ENDPOINTS>
      <ENDPOINT>
        <ID>8282</ID>
        <NAME>Name-it</NAME>
        <EXTENSIONS/>
        <URL>https://horizon.baz.example.com:5000/v3</URL>
        <INTERFACENAME>org.openstack.nova</INTERFACENAME>
        <ENDPOINT_MONITORED>Y</ENDPOINT_MONITORED>
      </ENDPOINT>
    </ENDPOINTS>
    <SCOPES>
      <SCOPE>EGI</SCOPE>
    </SCOPES>
    <EXTENSIONS/>
  </SERVICE_ENDPOINT>
</results>"""


class GOCDBTest(base.TestCase):
    def test_request_call(self):
        with mock.patch("requests.get") as m_requests:
            m_requests.return_value = mock.MagicMock()
            r = utils.find_in_gocdb("foo", "bar")
            m_requests.assert_called_once_with(
                "https://goc.egi.eu/gocdbpi/public/",
                params={"method": "get_service", "service_type": "bar"},
                verify=True,
                timeout=None,
            )
            assert {} == r

    def test_request_call_insecure(self):
        with mock.patch("requests.get") as m_requests:
            m_requests.return_value = mock.MagicMock()
            r = utils.find_in_gocdb("foo", "bar", insecure=True)
            m_requests.assert_called_once_with(
                "https://goc.egi.eu/gocdbpi/public/",
                params={"method": "get_service", "service_type": "bar"},
                verify=False,
                timeout=None,
            )
            assert {} == r

    def test_request_call_timeout(self):
        with mock.patch("requests.get") as m_requests:
            m_requests.return_value = mock.MagicMock()
            r = utils.find_in_gocdb("foo", "bar", timeout=1234)
            m_requests.assert_called_once_with(
                "https://goc.egi.eu/gocdbpi/public/",
                params={"method": "get_service", "service_type": "bar"},
                verify=True,
                timeout=1234,
            )
            assert {} == r

    def test_goc_non_200(self):
        with mock.patch("requests.get") as m_requests:
            r = mock.MagicMock()
            r.status_code = 404
            m_requests.return_value = r
            assert {} == utils.find_in_gocdb("foo", "bar")

    def test_goc_empty(self):
        with mock.patch("requests.get") as m_requests:
            r = mock.MagicMock()
            r.status_code = 200
            r.text = '<?xml version="1.0" encoding="UTF-8"?>' "<results/>"
            m_requests.return_value = r
            assert {} == utils.find_in_gocdb("foo", "bar")

    def test_goc_not_found(self):
        with mock.patch("requests.get") as m_requests:
            r = mock.MagicMock()
            r.status_code = 200
            r.text = sample_goc_response
            m_requests.return_value = r
            assert {} == utils.find_in_gocdb("foo", "bar")

    def test_goc_found_same_path(self):
        with mock.patch("requests.get") as m_requests:
            r = mock.MagicMock()
            r.status_code = 200
            r.text = sample_goc_response
            m_requests.return_value = r
            expected = {"gocdb_id": "1234G0", "site_name": "FOO-BAR-SITE"}
            assert expected == utils.find_in_gocdb(
                "https://keystone.example.com:5000/v2.0/", "bar"
            )

    def test_goc_found_similar_path(self):
        with mock.patch("requests.get") as m_requests:
            r = mock.MagicMock()
            r.status_code = 200
            r.text = sample_goc_response
            m_requests.return_value = r
            expected = {"gocdb_id": "1234G0", "site_name": "FOO-BAR-SITE"}
            assert expected == utils.find_in_gocdb(
                "https://keystone.example.com:5000/v2.0", "bar"
            )

    def test_goc_multiple_endpoints(self):
        with mock.patch("requests.get") as m_requests:
            r = mock.MagicMock()
            r.status_code = 200
            r.text = sample_goc_ep_response
            m_requests.return_value = r
            expected = {"gocdb_id": "00000G0", "site_name": "BAR-FOO-SITE"}
            assert expected == utils.find_in_gocdb(
                "https://horizon.baz.example.com:5000/v3", "bar"
            )



import json
import logging
import re
from urllib.parse import urljoin

import glanceclient
import novaclient.client
from cloud_info_provider import exceptions, glue, utils
from cloud_info_provider.providers import base
from keystoneauth1 import loading
from keystoneauth1.exceptions import http as http_exc
from keystoneauth1.loading import base as loading_base
from keystoneauth1.loading import session as loading_session
from novaclient.exceptions import Forbidden


class OpenStackProvider(base.BaseProvider):
    goc_service_type = "org.openstack.nova"
    interface_name = "org.openstack.nova"

    def setup_logging(self):
        super().setup_logging()
        # Remove info log messages from output
        external_logs = [
            "stevedore.extension",
            "requests",
            "urllib3",
            "novaclient",
            "keystoneauth",
            "keystoneclient",
        ]
        log_level = logging.DEBUG if self.opts.debug else logging.WARNING
        for log in external_logs:
            logging.getLogger(log).setLevel(log_level)

    def __init__(self, opts, **kwargs):
        super().__init__(opts, **kwargs)

        # NOTE(aloga): we do not want a project to be passed from the CLI,
        # as we will iterate over it for each configured VO and project.  We
        # have not added these arguments to the parser, but, since the plugin
        # is expecting them when parsing the arguments we need to set them to
        # None before calling the load_auth_from_argparse_arguments. However,
        # we may receive this in the "opts" namespace, therefore we do not set
        # it this is passed.
        if "os_project_name" not in opts:
            opts.os_project_name = None
            opts.os_tenant_name = None
        if "os_project_id" not in opts:
            opts.os_project_id = None
            opts.os_tenant_id = None
        self.project_id = None
        opts.os_auth_url = self.site_config["endpoint"]
        self.os_region = opts.os_region
        self.all_images = not opts.only_appdb_images

        # Select 'public', 'private' or 'all' (default) templates.
        self.select_flavors = opts.select_flavors

        self.flavor_properties = {}
        self.image_properties = {}
        for _opt in vars(self.opts):
            if _opt.startswith("property_flavor_") and not _opt.endswith("_value"):
                opts_k = vars(self.opts)[_opt]
                property_id = re.search(r"property_(\w+)", _opt).group(1)
                opts_v = vars(self.opts).get("_".join([_opt, "value"]), None)
                self.flavor_properties[property_id] = {"key": opts_k, "value": opts_v}
            elif _opt.startswith("property_image"):
                opts_k = vars(self.opts)[_opt]
                property_id = re.search(r"property_(\w+)", _opt).group(1)
                # keep the same structure, although we are not using the _value here
                self.image_properties[property_id] = {"key": opts_k, "value": None}
        self.last_working_auth = None
        self.exit_on_share_errors = self.opts.exit_on_share_errors

    def get_endpoint_id(self):
        return f"{self.site_config['endpoint']}_OpenStack_v3_oidc"

    def get_service_id(self):
        return f"{self.site_config['endpoint']}_cloud.compute"

    def get_manager_id(self):
        return f"{self.site_config['endpoint']}_cloud.compute_manager"

    def build_endpoint(self):
        return super().build_endpoint(
            implementor="OpenStack Foundation",
            implementation_name="OpenStack Nova",
            semantics="https://developer.openstack.org/api-ref/compute",
            # assuming this is correct,
            authentication="oidc",
        )

    def rescope_project(self, auth):
        """Switch to OS project whenever there is a change.

        It updates every OpenStack client used in case of new project.
        """
        project_id = auth["project_id"]
        region_name = auth.get("region_name", None)
        if self.project_id == project_id:
            return
        self.opts.os_project_id = project_id
        self.auth_plugin = loading.load_auth_from_argparse_arguments(self.opts)
        self.session = loading.load_session_from_argparse_arguments(
            self.opts, auth=self.auth_plugin
        )
        self.auth_plugin.invalidate()
        try:
            self.project_id = self.session.get_project_id()
        except http_exc.Unauthorized:
            # FIXME - this should be just reported in the glue and not break anything
            msg = "Could not authorize user in project '%s'" % project_id
            raise exceptions.OpenStackProviderException(msg)

        self.last_working_auth = auth

        # make sure the clients know about the change
        self.nova = novaclient.client.Client(
            2,
            session=self.session,
            region_name=region_name,
        )
        self.glance = glanceclient.Client(
            "2",
            session=self.session,
            region_name=region_name,
        )

    def build_instance_type(self, flavor, share):
        itype = glue.CloudComputingInstanceType(
            id=flavor.id,
            disk=flavor.disk,
            cpu=flavor.vcpus,
            name=flavor.name,
            ram=flavor.ram,
            platform="amd64",
        )
        itype.add_associated_object(share)
        itype.add_associated_object(self.endpoint)
        itype.add_associated_object(self.manager)
        # TODO add infiniband stuff, unclear where that lives in glue
        self.add_glue(itype)

        extra_properties = {}
        for property_id, opt in self.flavor_properties.items():
            v = flavor.get_keys().get(opt["key"])
            if v:
                opts_v = opt["value"]
                if opts_v:
                    extra_properties[property_id] = v == opts_v
                else:
                    extra_properties[property_id] = v
        if extra_properties.get("flavor_gpu_number"):
            acc = glue.CloudComputingVirtualAccelerator(
                id=f"{flavor.id}_gpu",
                name=f"{flavor.name}_gpu",
                type="GPU",
                number=extra_properties.get("flavor_gpu_number"),
                vendor=extra_properties.get("flavor_gpu_vendor", "UNKNOWN"),
                model=extra_properties.get("flavor_gpu_model", "UNKNOWN"),
            )
            acc.add_associated_object(itype)
            itype.add_associated_object(acc)
            self.add_glue(acc)
        return itype

    def build_share_instance_types(self, share):
        ram = []
        cpu = []
        add_all = self.select_flavors == "all"
        itypes = []
        for flavor in self.nova.flavors.list(detailed=True, is_public=None):
            add_pub = self.select_flavors == "public" and flavor.is_public
            add_priv = self.select_flavors == "private" and not flavor.is_public
            if not (add_all or add_pub or add_priv):
                continue
            itype = self.build_instance_type(flavor, share)
            ram.append(itype.ram)
            cpu.append(itype.cpu)
            itypes.append(itype)
        # update share max/min ram and cpu
        share.instance_max_ram = max(ram) if ram else 0
        share.instance_min_ram = min(ram) if ram else 0
        share.instance_max_cpu = max(cpu) if cpu else 0
        share.instance_min_cpu = min(cpu) if cpu else 0
        return itypes

    def build_image(self, image, share):
        image_descr = image.get(
            "vmcatcher_event_dc_description",
            image.get("vmcatcher_event_dc_title", "UNKNOWN"),
        )
        marketplace_url = image.get(
            "vmcatcher_event_ad_mpuri", image.get("marketplace")
        )

        other_info = {}
        try:
            extra_attrs = json.loads(image.get("APPLIANCE_ATTRIBUTES", "{}"))
        except ValueError:
            logging.warning(
                "Unexpected issue while getting json for '%s'",
                image.get("APPLIANCE_ATTRIBUTES", "{}"),
            )
            extra_attrs = {}
        # AppDB and cloud-info-api expect base_mpuri in OtherInfo
        # However atrope/cloudkeeper will push the attribute ad:base_mpuri instead
        # so we need to duplicate this for the time being
        if "ad:base_mpuri" in extra_attrs:
            other_info["base_mpuri"] = extra_attrs["ad:base_mpuri"]
        other_info.update(extra_attrs)

        if not marketplace_url:
            if self.all_images:
                link = urljoin(
                    self.glance.http_client.get_endpoint(), image.get("file")
                )
                marketplace_url = link
            else:
                return None

        glue_image = glue.CloudComputingImage(
            id=image["id"],
            name=image["name"],
            osName=image.get("os_distro", "UNKNOWN"),
            osVersion=image.get("os_version", "UNKNOWN"),
            osPlatform=image.get("architecture", "UNKNOWN"),
            description=image_descr,
            marketplace_url=marketplace_url,
            other_info=other_info,
        )
        glue_image.add_associated_object(share)
        glue_image.add_associated_object(self.endpoint)
        glue_image.add_associated_object(self.manager)
        self.add_glue(glue_image)
        # TODO add associated image objects (Image Network)
        return glue_image

    def build_share_images(self, share):
        image_list = []
        for image in self.glance.images.list(
            detailed=True, filters={"status": "active"}
        ):
            glue_img = self.build_image(image, share)
            if glue_img:
                image_list.append(glue_img)
        return image_list

    def build_share_quotas(self, share):
        """Return the quotas set for the current project."""
        quota_resources = [
            "instances",
            "cores",
            "ram",
            "floating_ips",
            "fixed_ips",
            "metadata_items",
            "injected_files",
            "injected_file_content_bytes",
            "injected_file_path_bytes",
            "key_pairs",
            "security_groups",
            "security_group_rules",
            "server_groups",
            "server_group_members",
        ]
        quotas = {}
        try:
            project_quotas = self.nova.quotas.get(self.project_id)
            for resource in quota_resources:
                try:
                    quotas[resource] = getattr(project_quotas, resource)
                except AttributeError:
                    pass
        except Forbidden:
            # Should we raise an error and make this mandatory?
            pass
        share.max_vm = quotas.get("instances", 0)
        share.other_info.update({"quotas": quotas})
        # also compute current instance count
        running_vm, halted_vm, suspended_vm = 0, 0, 0
        for s in self.nova.servers.list():
            if s.status == "SHUTOFF":
                halted_vm += 1
            elif s.status == "SUSPENDED":
                suspended_vm += 1
            else:  # this includes several cases (i.e. errors), but assume that's ok
                running_vm += 1
        share.running_vm = running_vm
        share.halted_vm = halted_vm
        share.suspended_vm = suspended_vm
        share.total_vm = running_vm + halted_vm + suspended_vm

    def build_shares(self):
        """Builds the share information for every VO"""
        share_objs = []
        rules = []
        total_vm, running_vm, halted_vm, suspended_vm = 0, 0, 0, 0
        max_cpu, min_cpu, max_ram, min_ram = 0, 0, 0, 0
        for vo in self.site_config["vos"]:
            try:
                self.rescope_project(vo["auth"])
            except exceptions.OpenStackProviderException as e:
                if self.exit_on_share_errors:
                    raise e
                else:
                    self.endpoint.health_state = "warn"
                    self.endpoint.health_state_info = str(e)
                    continue
            share_id = f"{self.get_endpoint_id()}_share_{vo['name']}_{self.project_id}"
            name = f"{vo['name']} - {self.project_id} share"
            description = f"Share in service {self.get_endpoint_id()} for VO {vo['name']} (Project {self.project_id})"
            access = self.auth_plugin.get_access(self.session)
            other_info = {
                "project_name": access.project_name,
                "project_domain_name": access.project_domain_name,
            }
            share = glue.Share(
                id=share_id,
                name=name,
                project_id=self.project_id,
                other_info=other_info,
                description=description,
            )
            share.add_associated_object(self.service)
            share.add_associated_object(self.endpoint)
            self.build_share_images(share)
            self.build_share_instance_types(share)
            self.build_share_quotas(share)
            max_ram = max(0, share.instance_max_ram)
            min_ram = min(0, share.instance_min_ram)
            max_cpu = max(0, share.instance_max_cpu)
            min_cpu = min(0, share.instance_min_cpu)
            self.add_glue(share)

            # policies
            rule = f"VO:{vo['name']}"
            rules.append(rule)
            mapping_policy = glue.MappingPolicy(id=f"{share_id}_Policy", rule=[rule])
            mapping_policy.add_associated_object(share)
            mapping_policy.add_association("PolicyUserDomain", vo["name"])
            self.add_glue(mapping_policy)

            running_vm += share.running_vm
            halted_vm += share.halted_vm
            suspended_vm += share.suspended_vm
            total_vm += share.total_vm

        # global policy
        access_policy = glue.AccessPolicy(
            id=f"{self.endpoint.id}_policy",
            rule=rules,
        )
        access_policy.add_associated_object(self.endpoint)
        self.add_glue(access_policy)

        # numbers about VMs in service or manager
        self.service.running_vm = running_vm
        self.service.halted_vm = halted_vm
        self.service.suspended_vm = suspended_vm
        self.service.total_vm = total_vm
        self.manager.instance_max_ram = max_ram
        self.manager.instance_min_ram = min_ram
        self.manager.instance_max_cpu = max_cpu
        self.manager.instance_min_cpu = min_cpu
        return share_objs

    def fetch(self):
        super().fetch()
        if self.last_working_auth:
            self.rescope_project(self.last_working_auth)
            self.endpoint.interface_version = self.nova.api_version.get_string()
            self.endpoint.implementation_version = (
                self.nova.versions.get_current().version
            )
        else:
            self.endpoint.health_state = "UNKNOWN"
            self.endpoint.health_state_info = "No working authentication configured"
        return self.objs

    @staticmethod
    def populate_parser(parser):
        """Populate the argparser 'parser' with the needed options."""
        base.BaseProvider.populate_parser(parser)

        plugins = loading_base.get_available_plugin_names()
        default_auth = "v3password"

        parser.add_argument(
            "--os-auth-type",
            "--os-auth-plugin",
            metavar="<name>",
            default=utils.env("OS_AUTH_TYPE", default=default_auth),
            choices=plugins,
            help="Authentication type to use, available "
            "types are: %s" % ", ".join(plugins),
        )

        # arguments come from session and plugins
        loading_session.register_argparse_arguments(parser)
        for plugin_name in plugins:
            plugin = loading_base.get_plugin_loader(plugin_name)
            # NOTE(aloga): we do not want a project to be passed from the
            # CLI, as we will iterate over it for each configured VO and
            # project. However, as the plugin is expecting them when
            # parsing the arguments we need to set them to None before
            # calling the load_auth_from_argparse_arguments method in the
            # __init__ method of this class.
            for opt in filter(
                lambda x: x.name not in ("project-name", "project-id"),
                plugin.get_options(),
            ):
                parser.add_argument(
                    *opt.argparse_args,
                    default=opt.argparse_default,
                    metavar="<auth-%s>" % opt.name,
                    help=opt.help,
                    dest="os_%s" % opt.dest.replace("-", "_"),
                )

        # somehow region is missing, so adding explicitly
        parser.add_argument("--os-region", default=None, help="OpenStack region to use")

        parser.add_argument(
            "--select-flavors",
            default="all",
            choices=["all", "public", "private"],
            help="Select all (default), public or private flavors.",
        )

        parser.add_argument(
            "--only-appdb-images",
            action="store_true",
            default=False,
            help=(
                "If set, only publish images with AppDB metadata, ignoring the others."
            ),
        )

        # PROPERTIES
        # If "property-<property>-value" is provided, the capability will only
        # be published when the given value matches the one in the flavor
        parser.add_argument(
            "--property-flavor-infiniband",
            metavar="PROPERTY_KEY",
            default="infiniband",
            help='Flavor"s property key for Infiniband support.',
        )

        parser.add_argument(
            "--property-flavor-infiniband-value",
            metavar="PROPERTY_VALUE",
            default="true",
            help=(
                "When Infiniband is supported, this option specifies the "
                "value to match."
            ),
        )

        parser.add_argument(
            "--property-flavor-gpu-number",
            metavar="PROPERTY_KEY",
            default="Accelerator:Number",
            help='Flavor"s property key pointing to number of GPUs.',
        )

        parser.add_argument(
            "--property-flavor-gpu-memory",
            metavar="PROPERTY_KEY",
            default="Accelerator:Memory",
            help='Flavor"s property key pointing to GPU memory.',
        )

        parser.add_argument(
            "--property-flavor-gpu-virt-type",
            metavar="PROPERTY_KEY",
            default="Accelerator:VirtualizationType",
            help='Flavor"s property key pointing to virtualization type.',
        )

        parser.add_argument(
            "--property-flavor-gpu-vendor",
            metavar="PROPERTY_KEY",
            default="Accelerator:Vendor",
            help='Flavor"s property key pointing to the GPU vendor.',
        )

        parser.add_argument(
            "--property-flavor-gpu-model",
            metavar="PROPERTY_KEY",
            default="Accelerator:Model",
            help='Flavor"s property key pointing to the GPU model.',
        )

        parser.add_argument(
            "--property-flavor-gpu-number",
            metavar="PROPERTY_KEY",
            default="Accelerator:Number",
            help='Flavor"s property key pointing to number of GPUs.',
        )

        parser.add_argument(
            "--property-image-gpu-driver",
            metavar="PROPERTY_KEY",
            default="gpu_driver",
            help='Image"s property key to specify the GPU driver version',
        )

        parser.add_argument(
            "--property-image-gpu-cuda",
            metavar="PROPERTY_KEY",
            default="gpu_cuda",
            help='Image"s property key to specify the CUDA toolkit version',
        )

        parser.add_argument(
            "--property-image-gpu-cudnn",
            metavar="PROPERTY_KEY",
            default="gpu_cudnn",
            help='Image"s property key to specify the cuDNN library version',
        )



import contextlib  # noqa

from cloud_info_provider import glue


def compare_glue(obj, glue_obj):
    for f in ("associations", "other_info"):
        if f not in obj:
            obj[f] = {}
    for field in glue_obj.__class__.model_fields:
        if field in ("creation_time", "validity"):
            continue
        v = getattr(glue_obj, field)
        if v is None:
            continue
        if isinstance(v, glue.BoolEnum):
            v = v.value
        if v != obj.get(field, None):
            print(f"Not matching {field} in object {obj} and {glue_obj.model_dump()}")
            return False
    # any extra fields??
    return True


@contextlib.contextmanager
def nested(*contexts):
    with contextlib.ExitStack() as stack:
        yield [stack.enter_context(c) for c in contexts]



import json
from datetime import datetime

from cloud_info_provider import glue
from cloud_info_provider.formatters import base


class GLUE21Json(base.BaseFormatter):
    def dump_glue_manager(self, mgr):
        return self.dump_glue_fields(
            mgr,
            {
                "ProductName": "product_name",
                "ProductVersion": "product_version",
                "HypervisorName": "hypervisor_name",
                "HypervisorVersion": "hypervisor_version",
                "TotalCPUs": "total_cpus",
                "TotalRAM": "total_ram",
                "InstanceMaxCPU": "instance_max_cpu",
                "InstanceMinCPU": "instance_min_cpu",
                "InstanceMaxRAM": "instance_max_ram",
                "InstanceMinRAM": "instance_min_ram",
                "NetworkVirtualizationType": "network_virtualization_type",
                "CPUVirtualizationType": "cpu_virtualization_type",
                "ManagerVirtualdiskFormat": "virtual_disk_format",
                "ManagerFailover": "failover",
                "ManagerLiveMigration": "live_migration",
                "ManagerVMBackupRestore": "vm_backup_restore",
            },
        )

    def dump_glue_endpoint(self, ept):
        return self.dump_glue_fields(
            ept,
            {
                "Capability": "capability",
                "QualityLevel": "quality_level",
                "InterfaceName": "interface_name",
                "InterfaceVersion": "interface_version",
                "HealthState": "health_state",
                "HealthStateInfo": "health_state_info",
                "ServingState": "serving_state",
                "Technology": "technology",
                "Implementor": "implementor",
                "ImplementationName": "implementation_name",
                "ImplementationVersion": "implementation_version",
                "DowntimeInfo": "downtime_info",
                "Semantics": "semantics",
                "Authentication": "authentication",
                "IssuerCA": "issuer_ca",
                "TrustedCA": "trusted_cas",
                "URL": "url",
            },
        )

    def dump_glue_image(self, img):
        return self.dump_glue_fields(
            img,
            {
                "MarketplaceURL": "marketplace_url",
                "OSPlatform": "osPlatform",
                "OSName": "osName",
                "OSVersion": "osVersion",
                "Description": "description",
                "AccessInfo": "access_info",
            },
        )

    def dump_glue_instance_type(self, itype):
        return self.dump_glue_fields(
            itype,
            {
                "Platform": "platform",
                "CPU": "cpu",
                "RAM": "ram",
                "Disk": "disk",
                "NetworkIn": "network_in",
                "NetworkOut": "network_out",
                "NetworkInfo": "network_info",
            },
        )

    def dump_glue_share(self, share):
        return self.dump_glue_fields(
            share,
            {
                "InstanceMaxCPU": "instance_max_cpu",
                "InstanceMaxRAM": "instance_max_ram",
                "SLA": "sla",
                "TotalVM": "total_vm",
                "RunningVM": "running_vm",
                "SuspendedVM": "suspended_vm",
                "HaltedVM": "halted_vm",
                "MaxVM": "max_vm",
                "NetworkInfo": "network_info",
                "DefaultNetworkType": "default_network_type",
                "PublicNetworkName": "public_network_name",
                "ProjectID": "project_id",
            },
        )

    def dump_glue_policy(self, policy):
        return self.dump_glue_fields(
            policy,
            {
                "Rule": "rule",
                "Scheme": "scheme",
            },
        )

    def dump_glue_accelerator(self, acc):
        return self.dump_glue_fields(
            acc,
            {
                "Type": "type",
                "Number": "number",
                "Vendor": "vendor",
                "Model": "model",
                "Version": "version",
                "ClockSpeed": "clock_speed",
                "Memory": "memory",
                "ComputeCapability": "compute_capability",
                "VirtualizationType": "virtualization_type",
            },
        )

    def dump_glue_service(self, svc):
        return self.dump_glue_fields(
            svc,
            {
                "Type": "type",
                "QualityLevel": "quality_level",
                "StatusInfo": "status_info",
                "ServiceAUP": "service_aup",
                "Complexity": "complexity",
                "Capability": "capability",
                "TotalVM": "total_vm",
                "RunningVM": "running_vm",
                "SuspendedVM": "suspended_vm",
                "HaltedVM": "halted_vm",
            },
        )

    def dump_glue_fields(self, obj, field_mapping):
        o = {}
        for name, field in field_mapping.items():
            v = getattr(obj, field, None)
            if v is not None:
                if isinstance(v, glue.BoolEnum):
                    o[name] = v.value
                elif isinstance(v, datetime):
                    o[name] = v.isoformat()
                else:
                    o[name] = v
        return o

    def dump_glue_object(self, obj):
        o = self.dump_glue_fields(
            obj,
            {
                "ID": "id",
                "Validity": "validity",
                "CreationTime": "creation_time",
                "Name": "name",
            },
        )
        if obj.other_info:
            o["OtherInfo"] = obj.other_info
        if obj.associations:
            o["Associations"] = obj.associations
        obj_mapping = {
            glue.CloudComputingService: self.dump_glue_service,
            glue.CloudComputingManager: self.dump_glue_manager,
            glue.CloudComputingEndpoint: self.dump_glue_endpoint,
            glue.CloudComputingImage: self.dump_glue_image,
            glue.CloudComputingInstanceType: self.dump_glue_instance_type,
            glue.Share: self.dump_glue_share,
            glue.MappingPolicy: self.dump_glue_policy,
            glue.AccessPolicy: self.dump_glue_policy,
            glue.CloudComputingVirtualAccelerator: self.dump_glue_accelerator,
        }
        f = obj_mapping.get(obj.__class__, None)
        if f:
            o.update(f(obj))
        return o

    def format(self, opts, glue):
        glue_output = {}
        for name, glue_objects in glue.items():
            glue_output[name] = [self.dump_glue_object(o) for o in glue_objects]

        return json.dumps(glue_output, default=str)



"""Test fixtures"""


class Data(object):
    @property
    def site_name(self):
        return "SITE_NAME"

    @property
    def endpoint_url(self):
        return "https://foo.example.org:5000/v3"

    @property
    def site_config(self):
        return {
            "gocdb": self.site_name,
            "endpoint": self.endpoint_url,
            "vos": [
                {"name": "foo1", "auth": {"project_id": "bar"}},
                {"name": "foo2", "auth": {"project_id": "baz"}},
            ],
        }


DATA = Data()


class OpenStackFakes(object):
    def __init__(self):
        class FakeObject(object):
            def __init__(self, **kwargs):
                self.d = kwargs
                for k, v in kwargs.items():
                    setattr(self, k, v)

            def get_keys(self):
                return vars(self)

            def get_dict(self):
                return self.d

        flavors = (
            {
                "id": "1",
                "name": "m1.foo",
                "ram": 10,
                "vcpus": 20,
                "is_public": True,
                "disk": 0,
                "ephemeral": 10,
                "infiniband": "true",
            },
            {
                "id": "2",
                "name": "m1 bar",
                "ram": 20,
                "vcpus": 30,
                "is_public": False,
                "disk": 10,
                "ephemeral": 10,
                "gpu_number": 23,
            },
            {
                "id": "3",
                "name": "m1.baz",
                "ram": 2,
                "vcpus": 3,
                "is_public": True,
                "disk": 5,
                "ephemeral": 10,
            },
        )

        self.flavors = [FakeObject(**f) for f in flavors]

        self.servers = [
            FakeObject(status=s)
            for s in (
                "SHUTOFF",
                "SUSPENDED",
                "SUSPENDED",
                "SUSPENDED",
                "SUSPENDED",
                "RUNNING",
                "RUNNING",
                "RUNNING",
            )
        ]

        # XXX add docker information
        # XXX some fake images should include more information from AppDB
        self.images = (
            {
                "name": "fooimage",
                "id": "foo.id",
                "metadata": {},
                "file": "v2/foo.id/file",
                "marketplace": "http://example.org/",
                "APPLIANCE_ATTRIBUTES": '{"ad:base_mpuri": "foobar"}',
                "gpu_driver": "driver-x",
                "gpu_cuda": "CUDA_x.y",
            },
            {
                "name": "barimage",
                "id": "bar id",
                "metadata": {},
                "file": "v2/bar id/file",
            },
            {
                "name": "bazimage",
                "id": "baz id",
                "docker_id": "sha1:xxxxxxxxxxxxxxxxxxxxxxxxxx",
                "docker_tag": "latest",
                "docker_name": "test/image",
                "file": "v2/baz id/file",
            },
        )

        self.quotas = FakeObject(
            instances=4,
            cores=5,
            ram=7,
            floating_ips=1,
            fixed_ips=0,
            metadata_items=4,
            injected_files=7,
            injected_file_content_bytes=8,
            injected_file_path_bytes=9,
            key_pairs=10,
            security_groups=1,
            security_group_rules=2,
            server_groups=4,
            server_group_members=5,
        )

        self.access = FakeObject(project_name="project", project_domain_name="default")

        self.version = FakeObject(version="vy.x")


OS_FAKES = OpenStackFakes()



import argparse

import mock
from cloud_info_provider import glue
from cloud_info_provider.exceptions import OpenStackProviderException
from cloud_info_provider.providers import openstack as os_provider
from cloud_info_provider.tests import base, data
from cloud_info_provider.tests import utils as utils
from keystoneauth1.exceptions import http as http_exc

FAKES = data.OS_FAKES


class OpenStackProviderOptionsTest(base.TestCase):
    def test_populate_parser(self):
        parser = argparse.ArgumentParser(conflict_handler="resolve")
        provider = os_provider.OpenStackProvider
        provider.populate_parser(parser)

        opts = parser.parse_args(
            [
                "--os-username",
                "foo",
                "--os-password",
                "bar",
                "--os-auth-url",
                "http://example.org:5000",
                "--insecure",
                "--only-appdb-images",
                "--select-flavors",
                "public",
                "--os-region",
                "North pole",
                "site_config",
            ]
        )

        assert opts.os_username == "foo"
        assert opts.os_password == "bar"
        assert opts.os_auth_url == "http://example.org:5000"
        assert opts.os_region == "North pole"
        assert opts.insecure
        assert opts.only_appdb_images
        assert opts.select_flavors == "public"


class OpenStackProviderAuthTest(base.TestCase):
    # Do not limit diff output on failures
    maxDiff = None

    def setUp(self):
        super().setUp()

        class FakeProvider(os_provider.OpenStackProvider):
            def __init__(self, opts):
                self.project_id = None
                self.os_region = None
                self.opts = mock.Mock()

        self.provider = FakeProvider(None)

    def test_rescope_simple(self):
        with utils.nested(
            mock.patch("keystoneauth1.loading.load_auth_from_argparse_arguments"),
            mock.patch("keystoneauth1.loading.load_session_from_argparse_arguments"),
        ) as (_, m_load_session):
            session = mock.Mock()
            session.get_project_id.return_value = "foo"
            m_load_session.return_value = session
            auth = {"project_id": "foo", "vo": "bar"}
            self.provider.rescope_project(auth)
            assert "foo" == self.provider.project_id
            assert auth == self.provider.last_working_auth

    def test_rescope_fails(self):
        with utils.nested(
            mock.patch("keystoneauth1.loading.load_auth_from_argparse_arguments"),
            mock.patch("keystoneauth1.loading.load_session_from_argparse_arguments"),
        ) as (_, m_load_session):
            session = mock.Mock()
            session.get_project_id.side_effect = http_exc.Unauthorized()
            m_load_session.return_value = session
            auth = {"project_id": "foo", "vo": "bar"}
            self.assertRaises(
                OpenStackProviderException, self.provider.rescope_project, auth
            )


class OpenStackProviderTest(base.TestCase):
    # Do not limit diff output on failures
    maxDiff = None

    def setUp(self):
        super().setUp()

        class FakeProvider(os_provider.OpenStackProvider):
            def rescope_project(self, auth):
                self.project_id = auth["project_id"]

            def __init__(self, opts):
                self.objs = {}
                self.nova = mock.Mock()
                self.nova.servers.list.return_value = FAKES.servers
                self.nova.flavors.list.return_value = FAKES.flavors
                self.nova.quotas.get.return_value = FAKES.quotas
                self.nova.versions.get_current.return_value = FAKES.version
                self.nova.api_version.get_string.return_value = "vx.y"
                self.glance = mock.Mock()
                self.glance.http_client.get_endpoint.return_value = (
                    "http://glance.example.org:9292/v2"
                )
                self.glance.images.list.return_value = FAKES.images
                self.session = mock.Mock()
                self.project_id = None
                self.session.get_project_id.return_value = "TEST_PROJECT_ID"
                self.auth_plugin = mock.MagicMock()
                self.auth_plugin.auth_url = data.DATA.endpoint_url
                self.auth_plugin.get_access.return_value = FAKES.access
                self.insecure = False
                self.os_region = None
                self.select_flavors = "all"
                self.all_images = False
                self.flavor_properties = {
                    "infiniband": {"key": "infiniband", "value": "true"},
                    "flavor_gpu_number": {"key": "gpu_number", "value": None},
                }
                self.image_properties = {
                    "image_gpu_driver": {"key": "gpu_driver", "value": "true"},
                    "image_gpu_cuda": {"key": "gpu_cuda", "value": "true"},
                }
                self.site_config = data.DATA.site_config
                self.add_glue(glue.CloudComputingService(id="foo", status_info="ok"))
                self.add_glue(
                    glue.CloudComputingEndpoint(
                        id="bar",
                        interface_name="iface",
                        url="url",
                    )
                )
                self.add_glue(glue.CloudComputingManager(id="baz"))
                self._goc_info = {data.DATA.endpoint_url: {"foo": "bar"}}
                self._ca_info = {data.DATA.endpoint_url: {"foo": "bar"}}
                self.last_working_auth = {"project_id": "foo"}

        self.provider = FakeProvider(None)

    def test_build_service(self):
        service = {
            "site_name": data.DATA.site_name,
            "name": f"Cloud Compute service at {data.DATA.site_name}",
            "other_info": {"foo": "bar"},
            "associations": {"AdminDomain": [data.DATA.site_name]},
            "status_info": f"https://argo.egi.eu/egi/report-status/Critical/SITES/{data.DATA.site_name}",
            "capability": [
                "executionmanagement.dynamicvmdeploy",
                "security.accounting",
            ],
            "type": "org.cloud.iaas",
            "id": "https://foo.example.org:5000/v3_cloud.compute",
            "quality_level": "production",
            "service_aup": "http://go.egi.eu/aup",
        }
        assert utils.compare_glue(service, self.provider.build_service())

    def test_provider_build_endpoint(self):
        endpoint = {
            "id": "https://foo.example.org:5000/v3_OpenStack_v3_oidc",
            "url": "https://foo.example.org:5000/v3",
            "name": "Cloud computing endpoint for https://foo.example.org:5000/v3_OpenStack_v3_oidc",
            "associations": {"CloudComputingService": ["foo"]},
            "capability": [],
            "quality_level": "production",
            "serving_state": "production",
            "interface_name": "org.openstack.nova",
            "implementor": "OpenStack Foundation",
            "implementation_name": "OpenStack Nova",
            "semantics": "https://developer.openstack.org/api-ref/compute",
            "health_state": "ok",
            "health_state_info": "Endpoint funtioning properly",
            "technology": "webservice",
            "authentication": "oidc",
            "downtime_info": "https://goc.egi.eu/portal/index.php?Page_Type=Downtimes_Calendar&site=SITE_NAME",
        }
        assert utils.compare_glue(endpoint, self.provider.build_endpoint())

    def test_build_image(self):
        share = glue.Share(id="share")
        image = self.provider.build_image(FAKES.images[0], share)
        assert utils.compare_glue(
            {
                "id": "foo.id",
                "name": "fooimage",
                "other_info": {"base_mpuri": "foobar", "ad:base_mpuri": "foobar"},
                "associations": {
                    "Share": ["share"],
                    "CloudComputingEndpoint": ["bar"],
                    "CloudComputingManager": ["baz"],
                },
                "marketplace_url": "http://example.org/",
                "osPlatform": "UNKNOWN",
                "osName": "UNKNOWN",
                "osVersion": "UNKNOWN",
                "description": "UNKNOWN",
                "access_info": "none",
            },
            image,
        )

    def test_build_images(self):
        self.provider.all_images = True
        share = glue.Share(id="share")
        images = self.provider.build_share_images(share)
        expected_images = {"bar id", "foo.id", "baz id"}
        assert expected_images == {i.id for i in images}
        self.provider.all_images = False
        images = self.provider.build_share_images(share)
        expected_images = {"foo.id"}
        assert expected_images == {i.id for i in images}

    def test_build_instance_type(self):
        share = glue.Share(id="share")
        itype = self.provider.build_instance_type(FAKES.flavors[0], share)
        assert utils.compare_glue(
            {
                "id": "1",
                "name": "m1.foo",
                "associations": {
                    "Share": ["share"],
                    "CloudComputingEndpoint": ["bar"],
                    "CloudComputingManager": ["baz"],
                },
                "platform": "amd64",
                "cpu": 20,
                "ram": 10,
                "disk": 0,
                "network_in": "UNKNOWN",
                "network_out": True,
            },
            itype,
        )

    def test_build_instance_type_gpu(self):
        share = glue.Share(id="share")
        itype = self.provider.build_instance_type(FAKES.flavors[1], share)
        assert utils.compare_glue(
            {
                "id": "2",
                "name": "m1 bar",
                "associations": {
                    "Share": ["share"],
                    "CloudComputingEndpoint": ["bar"],
                    "CloudComputingManager": ["baz"],
                    "CloudComputingVirtualAccelerator": ["2_gpu"],
                },
                "platform": "amd64",
                "cpu": 30,
                "ram": 20,
                "disk": 10,
                "network_in": "UNKNOWN",
                "network_out": True,
            },
            itype,
        )
        assert utils.compare_glue(
            {
                "id": "2_gpu",
                "name": "m1 bar_gpu",
                "associations": {"CloudComputingInstanceType": ["2"]},
                "compute_capability": [],
                "type": "GPU",
                "number": 23,
                "vendor": "UNKNOWN",
                "model": "UNKNOWN",
            },
            self.provider.get_first_obj("CloudComputingVirtualAccelerator"),
        )

    def test_build_instances(self):
        # all
        self.provider.select_flavors = "all"
        share = glue.Share(id="share")
        itypes = self.provider.build_share_instance_types(share)
        expected_types = {"1", "2", "3"}
        assert expected_types == {i.id for i in itypes}
        assert share.instance_max_ram == 20
        assert share.instance_min_ram == 2
        assert share.instance_max_cpu == 30
        assert share.instance_min_cpu == 3
        # public
        self.provider.select_flavors = "public"
        itypes = self.provider.build_share_instance_types(share)
        expected_types = {"1", "3"}
        assert expected_types == {i.id for i in itypes}
        assert share.instance_max_ram == 10
        assert share.instance_min_ram == 2
        assert share.instance_max_cpu == 20
        assert share.instance_min_cpu == 3
        # private
        self.provider.select_flavors = "private"
        itypes = self.provider.build_share_instance_types(share)
        expected_types = {"2"}
        assert expected_types == {i.id for i in itypes}
        assert share.instance_max_ram == 20
        assert share.instance_min_ram == 20
        assert share.instance_max_cpu == 30
        assert share.instance_min_cpu == 30

    def test_build_quotas(self):
        share = glue.Share(id="share")
        self.provider.build_share_quotas(share)
        assert share.running_vm == 3
        assert share.halted_vm == 1
        assert share.suspended_vm == 4
        assert share.total_vm == 8
        assert share.max_vm == 4
        assert share.other_info["quotas"] == FAKES.quotas.get_dict()

    def test_build_failing_share(self):
        def fail_rescope(auth):
            raise OpenStackProviderException("err")

        self.provider.exit_on_share_errors = False
        self.provider.rescope_project = fail_rescope
        self.provider.build_shares()
        shares = self.provider.get_objs("Share")
        assert shares == []

        self.provider.exit_on_share_errors = True
        self.assertRaises(OpenStackProviderException, self.provider.build_shares)

    def test_build_shares(self):
        self.provider.build_shares()
        # 2 shares and a global policy
        # For each share:
        # 1 image, 1 mapping policy, 3 flavors, 1 virtual accelerator
        shares = self.provider.get_objs("Share")
        assert {
            "https://foo.example.org:5000/v3_OpenStack_v3_oidc_share_foo1_bar",
            "https://foo.example.org:5000/v3_OpenStack_v3_oidc_share_foo2_baz",
        } == {s.id for s in shares}
        assert len(self.provider.get_objs("CloudComputingInstanceType")) == 6
        assert len(self.provider.get_objs("CloudComputingVirtualAccelerator")) == 2
        assert len(self.provider.get_objs("CloudComputingImage")) == 2
        assert len(self.provider.get_objs("MappingPolicy")) == 2
        assert len(self.provider.get_objs("AccessPolicy")) == 1

        bar_shares = [s for s in shares if s.project_id == "bar"][0]
        assert utils.compare_glue(
            {
                "id": "https://foo.example.org:5000/v3_OpenStack_v3_oidc_share_foo1_bar",
                "description": (
                    "Share in service "
                    "https://foo.example.org:5000/v3_OpenStack_v3_oidc"
                    " for VO foo1 (Project bar)"
                ),
                "name": "foo1 - bar share",
                "other_info": {
                    "project_name": "project",
                    "project_domain_name": "default",
                    "quotas": {
                        "instances": 4,
                        "cores": 5,
                        "ram": 7,
                        "floating_ips": 1,
                        "fixed_ips": 0,
                        "metadata_items": 4,
                        "injected_files": 7,
                        "injected_file_content_bytes": 8,
                        "injected_file_path_bytes": 9,
                        "key_pairs": 10,
                        "security_groups": 1,
                        "security_group_rules": 2,
                        "server_groups": 4,
                        "server_group_members": 5,
                    },
                },
                "associations": {
                    "CloudComputingService": ["foo"],
                    "CloudComputingEndpoint": ["bar"],
                },
                "instance_max_cpu": 30,
                "instance_max_ram": 20,
                "instance_min_cpu": 3,
                "instance_min_ram": 2,
                "total_vm": 8,
                "running_vm": 3,
                "suspended_vm": 4,
                "halted_vm": 1,
                "max_vm": 4,
                "project_id": "bar",
            },
            bar_shares,
        )

        # check policies, assume everything else is ok
        for o in self.provider.get_objs("MappingPolicy"):
            # either foo1 or foo2
            assert utils.compare_glue(
                {
                    "id": "https://foo.example.org:5000/v3_OpenStack_v3_oidc_share_foo2_baz_Policy",
                    "associations": {
                        "Share": [
                            "https://foo.example.org:5000/v3_OpenStack_v3_oidc_share_foo2_baz"
                        ],
                        "PolicyUserDomain": ["foo2"],
                    },
                    "scheme": "org.glite.standard",
                    "rule": ["VO:foo2"],
                },
                o,
            ) or utils.compare_glue(
                {
                    "id": "https://foo.example.org:5000/v3_OpenStack_v3_oidc_share_foo1_bar_Policy",
                    "associations": {
                        "Share": [
                            "https://foo.example.org:5000/v3_OpenStack_v3_oidc_share_foo1_bar"
                        ],
                        "PolicyUserDomain": ["foo1"],
                    },
                    "scheme": "org.glite.standard",
                    "rule": ["VO:foo1"],
                },
                o,
            )
        assert utils.compare_glue(
            {
                "id": "bar_policy",
                "associations": {"CloudComputingEndpoint": ["bar"]},
                "scheme": "org.glite.standard",
                "rule": ["VO:foo1", "VO:foo2"],
            },
            self.provider.get_first_obj("AccessPolicy"),
        )

    def test_fetch_working_auth(self):
        self.provider.fetch()
        assert self.provider.service.complexity == "endpointType=1,share=2"
        assert self.provider.endpoint.interface_version == "vx.y"
        assert self.provider.endpoint.implementation_version == "vy.x"

    def test_fetch_no_auth(self):
        self.provider.last_working_auth = None
        self.provider.fetch()
        assert self.provider.service.complexity == "endpointType=1,share=2"
        assert self.provider.endpoint.health_state == "UNKNOWN"
        assert (
            self.provider.endpoint.health_state_info
            == "No working authentication configured"
        )
        assert self.provider.endpoint.interface_version is None
        assert self.provider.endpoint.implementation_version is None



"""
Tests for the publishers
"""

from __future__ import print_function

import mock
from cloud_info_provider.publishers import stdout
from cloud_info_provider.tests import base


class StdOutPublisherTest(base.TestCase):
    def test_publish(self):
        publisher = stdout.StdOutPublisher(None)
        output = "foo"
        with mock.patch("builtins.print") as m_print:
            publisher.publish(output)
            m_print.assert_called_with(output)



import importlib.metadata

try:
    __version__ = importlib.metadata.version(__package__)
except importlib.metadata.PackageNotFoundError:
    # Package not installed in the environment (editable/dev checkout).
    # Fall back to a default version so importing the package doesn't fail
    # during local testing.
    __version__ = "0.0"



"""
StdOut Publisher

Just prints to stdout
"""

from __future__ import print_function

import json
from io import StringIO

from cloud_info_provider.publishers.base import BasePublisher


class StdOutPublisher(BasePublisher):
    @staticmethod
    def populate_parser(parser):
        """Populate the argparser 'parser' with the needed options."""
        pass

    def publish(self, output):
        print(output)


class JSONStdOutPublisher(BasePublisher):
    @staticmethod
    def populate_parser(parser):
        """Populate the argparser 'parser' with the needed options."""
        pass

    def publish(self, output):
        output_io = StringIO(output.replace("'", '"'))
        json_data = json.load(output_io)
        print(json.dumps(json_data, indent=4, sort_keys=True))



import logging

import yaml
from cloud_info_provider import glue
from cloud_info_provider.exceptions import CloudInfoException
from cloud_info_provider.providers import utils


class BaseProvider:
    goc_service_type = ""
    interface_name = ""

    def _load_site_config(self, config_file):
        with open(config_file, "r") as f:
            self.site_config = yaml.load(f.read(), Loader=yaml.SafeLoader)

        # Ensure the file has what we need
        for field in ("gocdb", "endpoint", "vos"):
            if field not in self.site_config:
                raise CloudInfoException("{field} not available in site config")

    def __init__(self, opts, **kwargs):
        self.opts = opts
        self.setup_logging()
        self._load_site_config(opts.site_config)
        self._goc_info = {}
        self._ca_info = {}
        self.objs = {}

    def _get_ca_info(self, url):
        if url not in self._ca_info:
            ca_info = utils.get_endpoint_ca_information(url, self.opts.insecure)
            self._ca_info[url] = ca_info
        return self._ca_info[url]

    def _get_goc_info(self, url):
        if url not in self._goc_info:
            # pylint: disable=no-member
            self._goc_info[url] = utils.find_in_gocdb(
                url, self.goc_service_type, self.opts.insecure, self.opts.timeout
            )
        return self._goc_info[url]

    def add_glue(self, o):
        class_name = o.__class__.__name__
        objs = self.objs.get(class_name, [])
        objs.append(o)
        self.objs[class_name] = objs

    def get_first_obj(self, obj_type):
        o = self.objs.get(obj_type, [None])
        return o[0]

    def get_objs(self, obj_type):
        return self.objs.get(obj_type, [])

    @property
    def service(self):
        return self.get_first_obj("CloudComputingService")

    @property
    def manager(self):
        return self.get_first_obj("CloudComputingManager")

    @property
    def endpoint(self):
        return self.get_first_obj("CloudComputingEndpoint")

    def fetch(self):
        self.build_service()
        self.build_manager()
        self.build_endpoint()
        self.build_shares()
        share_count = len(self.objs.get("Share", []))
        svc = self.service
        if svc:
            svc.complexity = f"endpointType=1,share={share_count}"
        return self.objs

    def get_service_id(self):
        return "service"

    def get_manager_id(self):
        return "manager"

    def get_endpoint_id(self):
        return "endpoint"

    def build_service(self, **kwargs):
        site_name = self.site_config["gocdb"]
        service_defaults = {
            "id": self.get_service_id(),
            "name": f"Cloud Compute service at {site_name}",
            "status_info": (
                f"https://argo.egi.eu/egi/report-status/Critical/SITES/{site_name}"
            ),
            "other_info": self._get_goc_info(self.site_config["endpoint"]),
        }
        service_defaults.update(kwargs)
        svc = glue.CloudComputingService(**service_defaults)
        svc.add_association("AdminDomain", site_name)
        self.add_glue(svc)
        return svc

    def build_manager(self, **kwargs):
        manager_defaults = {"id": self.get_manager_id()}
        manager_defaults.update(kwargs)
        mgr = glue.CloudComputingManager(**manager_defaults)
        mgr.add_associated_object(self.service)
        self.add_glue(mgr)
        return mgr

    def build_endpoint(self, **kwargs):
        ept_defaults = {
            "id": self.get_endpoint_id(),
            "name": f"Cloud computing endpoint for {self.get_endpoint_id()}",
            "interface_name": self.interface_name,
            "url": self.site_config["endpoint"],
            "health_state": "ok",
            "health_state_info": "Endpoint funtioning properly",
            "downtime_info": (
                "https://goc.egi.eu/portal/index.php?"
                f"Page_Type=Downtimes_Calendar&site={self.site_config['gocdb']}"
            ),
        }
        ca_info = self._get_ca_info(self.site_config["endpoint"])
        if "issuer" in ca_info:
            ept_defaults["issuer_ca"] = ca_info["issuer"]
        if "trusted_cas" in ca_info:
            ept_defaults["trusted_cas"] = ca_info["trusted_cas"]
        ept_defaults.update(kwargs)
        ept = glue.CloudComputingEndpoint(**ept_defaults)
        ept.add_associated_object(self.service)
        self.add_glue(ept)
        return ept

    def build_shares(self):
        pass

    def setup_logging(self):
        level = logging.DEBUG if self.opts.debug else logging.INFO
        logging.basicConfig(level=level)

    @staticmethod
    def populate_parser(parser):
        """Populate the argparser 'parser' with the needed options."""
        parser.add_argument(
            "site_config",
            help="YAML file with site configuration (as in fedcloud-catchall-ops)",
        )



import mock
from cloud_info_provider import core
from cloud_info_provider.tests import base


class CoreOptionsTest(base.TestCase):
    def test_populate_parser(self):
        providers = {"foo": mock.Mock()}
        formatters = ["foobar"]
        publishers = {"boom": mock.Mock()}
        parser = core.get_parser(providers, formatters, publishers)
        opts = parser.parse_args(
            [
                "--middleware",
                "foo",
                "--format",
                "foobar",
                "--publisher",
                "boom",
                "--debug",
            ]
        )
        assert opts.middleware == "foo"
        assert opts.format == "foobar"
        assert opts.publisher == "boom"
        assert opts.debug

    def test_populate_parser_defaults(self):
        parser = core.get_parser({}, [], {})
        opts = parser.parse_args([])
        assert opts.middleware == "openstack"
        assert opts.format == "glue21json"
        assert opts.publisher == "stdout"
        assert not opts.debug
